[From nobody Tue Sep  5 14:23:59 2006
Subject: Locking fixes, receive fixes.
From: David Woodhouse &lt;dwmw2@infradead.org&gt;
To: theo@crazygreek.co.uk
Cc: chan_bluetooth@lists.crazygreek.co.uk
Content-Type: text/plain
Message-Id: &lt;1111634624.4533.40.camel@localhost.localdomain&gt;
Mime-Version: 1.0
X-Mailer: Evolution 2.2.1.1 (2.2.1.1-1) 
Date: Thu, 24 Mar 2005 03:23:49 +0000
X-Evolution-Format: text/plain
X-Evolution-Account: 0
X-Evolution-Transport: smtp://dwmw2; auth=CRAM-MD5@213.146.154.40:587/;
	use_ssl=when-possible
X-Evolution-Fcc: email://0/outgoing
X-Evolution-Source: imap://dwmw2@pentafluge.infradead.org/
Content-Transfer-Encoding: 7bit

With the attached patch, chan_bluetooth works OK in full duplex mode
with Asterisk CVS, with both an Ericsson T630 and a Plantronics M3000
headset. I fixed some lock ordering bugs which led to deadlock, but I'm
still dubious about the locking, and there still seems to be a race with
starting the SCO thread -- if the dial fails and we hang up before it
manages to actually start up, we don't catch it. But it's usable like
this...

I implemented a simple way to initiate a call by using the 'redial'
functionality of the headset, but it would probably be better to do
something along the lines of the OSS and ALSA 'console' drivers. 

--- chan_bluetooth.c.orig	2005-03-24 02:51:42.026484034 +0000
+++ chan_bluetooth.c	2005-03-24 03:22:27.456879897 +0000
@@ -96,7 +96,6 @@
 #include &lt;asterisk/lock.h&gt;
 #include &lt;asterisk/utils.h&gt;
 #include &lt;asterisk/channel.h&gt;
-#include &lt;asterisk/channel_pvt.h&gt;
 #include &lt;asterisk/config.h&gt;
 #include &lt;asterisk/logger.h&gt;
 #include &lt;asterisk/module.h&gt;
@@ -221,6 +220,7 @@ struct blt_dev {
   char rd_buff[BLT_RDBUFF_MAX];     /* RFCOMM input buffer */
   int rd_buff_pos;                  /* RFCOMM input buffer position */
   int ready;                        /* 1 When ready */
+  char *context;
 
   /* AG mode */
   char last_ok_cmd[BLT_RDBUFF_MAX];        /* Runtime[AG]: Last AT command that was OK */
@@ -306,6 +306,33 @@ static struct sched_context * sched = NU
 
 /* ---------------------------------- */
 
+#if ASTERISK_VERSION_NUM &lt;= 010107
+#include &lt;asterisk/channel_pvt.h&gt;
+#define tech_pvt pvt-&gt;pvt
+#else /* CVS. FIXME: Version number */
+static struct ast_channel *blt_request(const char *type, int format, void *data, int *cause);
+static int blt_hangup(struct ast_channel *c);
+static int blt_answer(struct ast_channel *c);
+static struct ast_frame *blt_read(struct ast_channel *chan);
+static int blt_call(struct ast_channel *c, char *dest, int timeout);
+static int blt_write(struct ast_channel *chan, struct ast_frame *f);
+static int blt_indicate(struct ast_channel *chan, int cond);
+
+static const struct ast_channel_tech blt_tech = {
+	.type = BLT_CHAN_NAME,
+	.description = &quot;Bluetooth Channel Driver&quot;,
+	.capabilities = BLUETOOTH_FORMAT,
+	.requester = blt_request,
+	.hangup = blt_hangup,
+	.answer = blt_answer,
+	.read = blt_read,
+	.call = blt_call,
+	.write = blt_write,
+	.indicate = blt_indicate,
+};
+#endif
+/* ---------------------------------- */
+
 static const char *
 role2str(blt_role_t role)
 {
@@ -562,7 +589,6 @@ sco_thread(void * data)
     if (res == 0)
       continue;
 
-    ast_mutex_lock(&amp;(dev-&gt;sco_lock));
 
     if (pfd[0].revents &amp; POLLIN) {
 
@@ -570,16 +596,16 @@ sco_thread(void * data)
 
       if (len) {
         ast_mutex_lock(&amp;(dev-&gt;lock));
+        ast_mutex_lock(&amp;(dev-&gt;sco_lock));
         set_buffer(dev-&gt;sco_buf_in,  buf, BUFLEN, &amp;in_pos,  len);
         get_buffer(buf, dev-&gt;sco_buf_out, BUFLEN, &amp;out_pos, len);
         write(dev-&gt;sco, buf, len);
+        ast_mutex_unlock(&amp;(dev-&gt;sco_lock));
+        ast_mutex_unlock(&amp;(dev-&gt;lock));
         if (dev-&gt;owner &amp;&amp; dev-&gt;owner-&gt;_state == AST_STATE_UP)
           write(dev-&gt;sco_pipe[1], &amp;c, 1);
-        ast_mutex_unlock(&amp;(dev-&gt;lock));
       }
 
-      ast_mutex_unlock(&amp;(dev-&gt;sco_lock));
-
     } else if (pfd[0].revents) {
 
       int e = sock_err(pfd[0].fd);
@@ -710,7 +736,7 @@ answer(blt_dev_t * dev)
 static int 
 blt_write(struct ast_channel * ast, struct ast_frame * frame)
 {
-  blt_dev_t * dev = ast-&gt;pvt-&gt;pvt; 
+  blt_dev_t * dev = ast-&gt;tech_pvt; 
 
   /* Write a frame of (presumably voice) data */
 
@@ -739,7 +765,7 @@ blt_write(struct ast_channel * ast, stru
 static struct ast_frame *
 blt_read(struct ast_channel * ast)
 {
-  blt_dev_t * dev = ast-&gt;pvt-&gt;pvt;
+  blt_dev_t * dev = ast-&gt;tech_pvt;
   char c = 1;
   int len;
 
@@ -753,11 +779,12 @@ blt_read(struct ast_channel * ast)
   dev-&gt;fr.mallocd = 0;
   dev-&gt;fr.delivery.tv_sec = 0;
   dev-&gt;fr.delivery.tv_usec = 0;
-
+  read(dev-&gt;sco_pipe[0], &amp;c, 1);
   ast_mutex_lock(&amp;(dev-&gt;sco_lock));
   dev-&gt;sco_sending = 1;
-  read(dev-&gt;sco_pipe[0], &amp;c, 1);
   len = get_buffer(dev-&gt;buf, dev-&gt;sco_buf_in, BUFLEN, &amp;(dev-&gt;sco_pos_in), 48);
+  /* Er, get_buffer() always returns zero. */
+  len = 48;
   ast_mutex_unlock(&amp;(dev-&gt;sco_lock));
 
   dev-&gt;fr.data = dev-&gt;buf;
@@ -872,7 +899,7 @@ ring_hs(blt_dev_t * dev)
 static int
 blt_call(struct ast_channel * ast, char * dest, int timeout)
 {
-  blt_dev_t * dev = ast-&gt;pvt-&gt;pvt;
+  blt_dev_t * dev = ast-&gt;tech_pvt;
 
   if ((ast-&gt;_state != AST_STATE_DOWN) &amp;&amp; (ast-&gt;_state != AST_STATE_RESERVED)) {
     ast_log(LOG_WARNING, &quot;blt_call called on %s, neither down nor reserved\n&quot;, ast-&gt;name);
@@ -927,11 +954,11 @@ blt_call(struct ast_channel * ast, char 
 static int 
 blt_hangup(struct ast_channel * ast)
 {
-  blt_dev_t * dev = ast-&gt;pvt-&gt;pvt;
+  blt_dev_t * dev = ast-&gt;tech_pvt;
 
   ast_log(LOG_DEBUG, &quot;blt_hangup(%s)\n&quot;, ast-&gt;name);
 
-  if (!ast-&gt;pvt-&gt;pvt) {
+  if (!ast-&gt;tech_pvt) {
     ast_log(LOG_WARNING, &quot;Asked to hangup channel not connected\n&quot;);
     return 0;
   }
@@ -976,7 +1003,7 @@ blt_hangup(struct ast_channel * ast)
   if (dev-&gt;status == BLT_STATUS_IN_CALL || dev-&gt;status == BLT_STATUS_RINGING)
     dev-&gt;status = BLT_STATUS_READY;
 
-  ast-&gt;pvt-&gt;pvt = NULL;
+  ast-&gt;tech_pvt = NULL;
   dev-&gt;owner = NULL;
   ast_mutex_unlock(&amp;(dev-&gt;lock));
   ast_setstate(ast, AST_STATE_DOWN);
@@ -1003,7 +1030,7 @@ blt_indicate(struct ast_channel * c, int
 static int
 blt_answer(struct ast_channel * ast)
 {
-  blt_dev_t * dev = ast-&gt;pvt-&gt;pvt;
+  blt_dev_t * dev = ast-&gt;tech_pvt;
 
   ast_mutex_lock(&amp;dev-&gt;lock);
 
@@ -1041,8 +1068,8 @@ blt_new(blt_dev_t * dev, int state, cons
   // ast-&gt;fds[0] = socket(PF_BLUETOOTH, SOCK_SEQPACKET, BTPROTO_SCO);
 
   ast-&gt;nativeformats       = BLUETOOTH_FORMAT;
-  ast-&gt;pvt-&gt;rawreadformat  = BLUETOOTH_FORMAT;
-  ast-&gt;pvt-&gt;rawwriteformat = BLUETOOTH_FORMAT;
+  //  ast-&gt;pvt-&gt;rawreadformat  = BLUETOOTH_FORMAT;
+  //  ast-&gt;pvt-&gt;rawwriteformat = BLUETOOTH_FORMAT;
   ast-&gt;writeformat         = BLUETOOTH_FORMAT;
   ast-&gt;readformat          = BLUETOOTH_FORMAT;
 
@@ -1050,15 +1077,17 @@ blt_new(blt_dev_t * dev, int state, cons
 
   ast-&gt;type = BLT_CHAN_NAME;
 
-  ast-&gt;pvt-&gt;pvt = dev;
-
+  ast-&gt;tech_pvt = dev;
+#if ASTERISK_VERSION_NUM &gt; 010107
+  ast-&gt;tech = &amp;blt_tech;
+#else
   ast-&gt;pvt-&gt;call     = blt_call;
   ast-&gt;pvt-&gt;indicate = blt_indicate;
   ast-&gt;pvt-&gt;hangup   = blt_hangup;
   ast-&gt;pvt-&gt;read     = blt_read;
   ast-&gt;pvt-&gt;write    = blt_write;
   ast-&gt;pvt-&gt;answer   = blt_answer;
-
+#endif
   strncpy(ast-&gt;context, context, sizeof(ast-&gt;context)-1);
   strncpy(ast-&gt;exten,   number,  sizeof(ast-&gt;exten) - 1);
 
@@ -1088,8 +1117,10 @@ blt_new(blt_dev_t * dev, int state, cons
 static struct ast_channel *
 #if (ASTERISK_VERSION_NUM &lt; 010100)
 blt_request(char * type, int format, void * local_data)
-#else
+#elif (ASTERISK_VERSION_NUM &lt;= 010107)
 blt_request(const char * type, int format, void * local_data)
+#else
+blt_request(const char * type, int format, void * local_data, int *cause)
 #endif
 {
   char * data = (char*)local_data;
@@ -1420,7 +1451,7 @@ atcmd_dial_execute(blt_dev_t * dev, cons
 
   sco_start(dev, -1);
 
-  if (blt_new(dev, AST_STATE_UP, &quot;bluetooth&quot;, number) == NULL) {
+  if (blt_new(dev, AST_STATE_UP, dev-&gt;context, number) == NULL) {
     sco_stop(dev);
   }
 
@@ -1429,6 +1460,11 @@ atcmd_dial_execute(blt_dev_t * dev, cons
   return 0;
 }
 
+static int atcmd_bldn_execute(blt_dev_t * dev, const char *data)
+{
+	return atcmd_dial_execute(dev, &quot;bldn;&quot;);
+}
+
 /* Answer */
 
 static int
@@ -1564,6 +1600,7 @@ atcmd_list[] = 
   { &quot;+CSCS&quot;, atcmd_cscs_set, NULL,            NULL,                 NULL,             NULL },
   { &quot;*EIPS&quot;, atcmd_eips_set, NULL,            NULL,                 NULL,             NULL },
   { &quot;+VGS&quot;,  atcmd_vgs_set,  NULL,            NULL,                 NULL,             NULL },
+  { &quot;+BLDN&quot;, NULL,           NULL,            atcmd_bldn_execute,   NULL,             NULL },
 };
 
 #define ATCMD_LIST_LEN (sizeof(atcmd_list) / sizeof(blt_atcb_t))
@@ -2135,7 +2172,7 @@ process_rfcomm_cmd(blt_dev_t * dev, char
 
   cmd += 2;
 
-  // Don't forget 'AT' on it's own is OK.
+  // Don't forget 'AT' on its own is OK.
 
   if (strlen(cmd) == 0) {
     send_atcmd_ok(dev, fullcmd);
@@ -2761,6 +2798,11 @@ blt_parse_config(void)
       if ((str = ast_variable_retrieve(cfg, cat, &quot;autoconnect&quot;)) != NULL)
         device-&gt;autoconnect = (strcasecmp(str, &quot;yes&quot;) == 0 || strcmp(str, &quot;1&quot;) == 0) ? 1 : 0;
 
+      if ((str = ast_variable_retrieve(cfg, cat, &quot;context&quot;)) != NULL)
+	device-&gt;context = str;
+      else
+	device-&gt;context = &quot;bluetooth&quot;;
+
       device-&gt;next = iface_head;
       iface_head = device;
       ifcount++;
@@ -2979,7 +3021,11 @@ static int
 __unload_module(void)
 {
 
+#if ASTERISK_VERSION_NUM &lt;= 010107
   ast_channel_unregister(BLT_CHAN_NAME);
+#else
+  ast_channel_unregister(&amp;blt_tech);
+#endif
 
   if (monitor_thread != AST_PTHREADT_NULL) {
 
@@ -3077,7 +3123,11 @@ load_module()
   if (restart_monitor() != 0)
     return -1;
 
+#if ASTERISK_VERSION_NUM &lt;= 010107
   if (ast_channel_register(BLT_CHAN_NAME, &quot;Bluetooth Driver&quot;, BLUETOOTH_FORMAT, blt_request)) {
+#else
+  if (ast_channel_register(&amp;blt_tech)) {
+#endif
     ast_log(LOG_ERROR, &quot;Unable to register channel class BTL\n&quot;);
     __unload_module();
     return -1;


-- 
dwmw2
]